/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package sample.tswebclient.cassandra;

import com.datastax.driver.core.BoundStatement;
import com.datastax.driver.core.Cluster;
import com.datastax.driver.core.ColumnDefinitions;
import com.datastax.driver.core.Host;
import com.datastax.driver.core.Metadata;
import com.datastax.driver.core.PreparedStatement;
import com.datastax.driver.core.ResultSet;
import com.datastax.driver.core.Row;
import com.datastax.driver.core.Session;
import com.datastax.driver.core.utils.Bytes;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.Arrays;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * CREATE KEYSPACE applications WITH REPLICATION = { 'class' : 'SimpleStrategy',
 * 'replication_factor' : 1 };
 *
 * USE applications;
 *
 * CREATE TABLE applications (key text PRIMARY KEY, name text, type int, 
 * icon blob, description text, path text, rdp blob, parent text);
 *
 * CREATE INDEX parents ON applications(parent);
 *
 *
 * @author ricky
 */
@javax.inject.Named
@javax.enterprise.context.ApplicationScoped
public class CassandraManager implements AutoCloseable {

    private static final Logger logger = LoggerFactory.getLogger(CassandraManager.class);

    public enum ChildrenReadOption {

        NONE, ONLY_NAMES, FULL, RECURSIVE,
    }

    private Cluster cluster = null;
    private Session session = null;
    private PreparedStatement insertApplication = null;
    private PreparedStatement deleteApplication = null;
    private PreparedStatement selectAll = null;
    private PreparedStatement selectKeyName = null;
    // TODO: put some configuration here
    private String node = "localhost";
    private int port = 9042;
    
    public CassandraManager() {
    }
    
    public boolean isInitialize() {
        return session != null;
    }
    
    @PostConstruct
    public synchronized void init() {
        cluster = Cluster.builder()
                .addContactPoint(node)
                .withPort(port)
                .build();
        Metadata metadata = cluster.getMetadata();
        logger.debug("Connected to cluster: {}", metadata.getClusterName());
        for (Host host : metadata.getAllHosts()) {
            logger.debug("Datatacenter: {}; Host: {}; Rack: {}",
                    host.getDatacenter(), host.getAddress(), host.getRack());
        }
        session = cluster.connect();
        // prepare statements
        insertApplication = session.prepare(
                "INSERT INTO applications.applications(key, name, type, icon, description, path, rdp, parent) VALUES(?, ?, ?, ?, ?, ?, ?, ?);");
        deleteApplication = session.prepare(
                "DELETE FROM applications.applications WHERE key = ?;");
        selectAll = session.prepare(
                "SELECT key, name, type, icon, writetime(icon) as icontimestamp, "
                + "description, path, rdp, writetime(rdp) as rdptimestamp, parent "
                + "FROM applications.applications "
                + "WHERE key = ?;");
        selectKeyName = session.prepare("SELECT key, name FROM applications.applications WHERE parent = ?;");
    }

    // AutoCloseable
    @Override
    @PreDestroy
    public void close() {
        session.close();
        cluster.close();
    }

    private Application fillApplication(Application app, ColumnDefinitions colDefs, Row row) {
        for (ColumnDefinitions.Definition def : colDefs.asList()) {
            switch (def.getName()) {
                case "name":
                    app.setName(row.getString("name"));
                    break;
                case "type":
                    app.setType(Application.Type.values()[row.getInt("type")]);
                    break;
                case "description":
                    app.setDescription(row.getString("description"));
                    break;
                case "icon":
                    if (row.getBytes("icon") != null) {
                        app.setIcon(Bytes.getArray(row.getBytes("icon")));
                    }
                    break;
                case "icontimestamp":
                    // set timestamp in millis (nit micro)
                    app.setWriteIconTimestamp(row.getLong("icontimestamp")/1000L);
                    break;
                case "path":
                    app.setPath(row.getString("path"));
                    break;
                case "rdp":
                    if (row.getBytes("rdp") != null) {
                        app.setRdp(Bytes.getArray(row.getBytes("rdp")));
                    }
                    break;
                case "rdptimestamp":
                    // set timestamp in millis (nit micro)
                    app.setWriteRdpTimestamp(row.getLong("rdptimestamp")/1000L);
                    break;
            }
        }
        return app;
    }

    //
    // PUBLIC METHODS
    //
    public Application readApplication(String key, ChildrenReadOption readOptions) {
        logger.debug("init {} {} {}", key, readOptions);
        StringBuilder sql = new StringBuilder();
        BoundStatement boundStatement = new BoundStatement(selectAll);
        ResultSet results = session.execute(boundStatement.bind(key));
        Row row = results.one();
        Application app = null;
        if (row != null) {
            app = new Application(row.getString("key"));
            fillApplication(app, results.getColumnDefinitions(), row);
            // read the children if needed
            if (readOptions.compareTo(ChildrenReadOption.NONE) > 0
                    && Application.Type.FOLDER.equals(app.getType())) {
                // read the key and the name
                sql = new StringBuilder();
                boundStatement = new BoundStatement(selectKeyName);
                results = session.execute(boundStatement.bind(key));
                for (Row r : results) {
                    if (readOptions.compareTo(ChildrenReadOption.ONLY_NAMES) > 0) {
                        // I have to read the children full or recursive
                        app.getChildren().add(
                                readApplication(r.getString("key"),
                                        readOptions.equals(ChildrenReadOption.RECURSIVE) ? ChildrenReadOption.RECURSIVE : ChildrenReadOption.NONE));
                    } else {
                        // only read the names of the children
                        Application child = new Application(r.getString("key"));
                        child.setName(r.getString("name"));
                        app.getChildren().add(child);
                    }
                }
            }
        }
        return app;
    }

    public void createApplication(Application app) {
        logger.debug("init {}", app);
        BoundStatement boundStatement = new BoundStatement(insertApplication);
        ByteBuffer buf = ByteBuffer.wrap(app.getIcon());
        System.err.println("ORIG: " + app.getIcon().length);
        System.err.println("BUFF: " + Bytes.getArray(buf).length);
        System.err.println("Equals: "  + Arrays.equals(app.getIcon(), Bytes.getArray(buf)));
        boundStatement.bind(app.getKey(), 
                app.getName(), 
                app.getType().ordinal(), 
                app.getIcon() == null? null : ByteBuffer.wrap(app.getIcon()),
                app.getDescription(), 
                app.getPath(), 
                app.getRdp() == null? null : ByteBuffer.wrap(app.getRdp()), 
                app.getParentKey());
        session.execute(boundStatement);
    }

    public void deleteApplication(String key) throws Exception {
        logger.debug("init {}", key);
        BoundStatement boundStatement = new BoundStatement(deleteApplication);
        boundStatement.bind(key);
        session.execute(boundStatement);
    }

    static private byte[] readFile(String name) throws FileNotFoundException, IOException {
        byte[] file = new byte[4096];
        InputStream input = new FileInputStream("/home/ricky/NetBeansProjects/tswebclient2/extras/" + name);
        int length = input.read(file);
        if (length > 0 && length < file.length) {
            file = Arrays.copyOfRange(file, 0, length);
        } else {
            throw new IOException("File too big!!!");
        }
        return file;
    }

    public static void main(String[] args) throws Exception {
        try (CassandraManager manager = new CassandraManager()) {
            manager.init();
            // initialization
            Application root = Application.root();
            // internet directory
            Application internet = new Application(root, "Internet", Application.Type.FOLDER,
                    null, "Example application folder", null, null);
            Application skype = new Application(internet, "Skype", Application.Type.APPLICATION,
                    readFile("skype.png"), "Skype application.", null, null);
            Application pidgin = new Application(internet, "Pidgin", Application.Type.APPLICATION,
                    readFile("pidgin.png"), "Pidgin application.", null, null);
            // util directory
            Application util = new Application(root, "Util", Application.Type.FOLDER,
                    null, "Util folder", null, null);
            Application paint = new Application(util, "Paint", Application.Type.APPLICATION,
                    readFile("paint.png"),
                    "Creates and edits drawings, and displays and edits scanned photos.", null,
                    readFile("paint.rdp"));
            Application notepad = new Application(util, "Notepad", Application.Type.APPLICATION,
                    readFile("notepad.png"),
                    "Creates and edits text files using basic text formatting.", null,
                    readFile("notepad.rdp"));
            Application wordpad = new Application(util, "WordPad", Application.Type.APPLICATION,
                    readFile("wordpad.png"),
                    "Creates and edits text documents with complex formatting.", null,
                    readFile("wordpad.rdp"));
            // Office directory
            Application office = new Application(root, "Office", Application.Type.FOLDER,
                    null, "Folder for office applications.", null, null);
            Application acroread = new Application(office, "Acrobat Reader", Application.Type.APPLICATION,
                    readFile("acroread.png"),
                    "Easily view, print, and collaborate on PDF files with free Adobe Reader 9 software.", null,
                    null);
            Application openoofice = new Application(office, "OpenOffice", Application.Type.FOLDER,
                    null, "OpenOffice software suite", null, null);
            Application oowriter = new Application(openoofice, "Writer", Application.Type.APPLICATION,
                    readFile("openofficeorg3-writer.png"),
                    "Create and edit text and graphics in letters, reports, documents and Web pages by using Writer.", null,
                    null);
            Application oocalc = new Application(openoofice, "Calc", Application.Type.APPLICATION,
                    readFile("openofficeorg3-calc.png"),
                    "Perform calculation, analyze information and manage lists in spreadsheets by using Calc.", null,
                    null);
            Application microsoftoffice = new Application(office, "Microsoft Office", Application.Type.FOLDER,
                    null, "Microsoft Office software suite", null, null);
            // Data Bases
            Application ddbb = new Application(root, "DDBB", Application.Type.FOLDER,
                    null, "Custom DataBase programs.", null, null);

            // Test
            Application test = new Application(Application.root(), "Applications", Application.Type.FOLDER,
                    null, "Descripción de applications", null, null);
            
//            manager.createApplication(root);
//            manager.createApplication(internet);
//            manager.createApplication(skype);
//            manager.createApplication(pidgin);
//            manager.createApplication(util);
//            manager.createApplication(paint);
//            manager.createApplication(notepad);
//            manager.createApplication(wordpad);
//            manager.createApplication(office);
//            manager.createApplication(acroread);
//            manager.createApplication(openoofice);
//            manager.createApplication(oowriter);
//            manager.createApplication(oocalc);
//            manager.createApplication(microsoftoffice);
//            manager.createApplication(ddbb);
//            manager.createApplication(test);
//            manager.deleteApplication(test.getKey());

            System.err.println("read root");
            root = manager.readApplication(root.getKey(), ChildrenReadOption.RECURSIVE);
            System.err.println("read internet");
            internet = manager.readApplication(internet.getKey(), ChildrenReadOption.ONLY_NAMES);
            System.err.println("read root");
            pidgin = manager.readApplication(pidgin.getKey(), ChildrenReadOption.NONE);

            System.err.print(root.printPath());
            System.err.println(manager.readApplication(test.getKey(), ChildrenReadOption.NONE));
        }
    }
}
